import datetime
# This assumes that you've only got utc times in your database.  Please don't put anything but utc in your database.  I'll cry.

def list_frequency(mylist):
    """ Turn a sequence with duplicates into a list of tuples matching item to frequency  """
    out = []
    if len(mylist) < 1:
        return out
    counter = {}
    for i in set(mylist):
        counter[i] = mylist.count(i)
    for x in counter:
        out.append((x, counter[x]))
    return sorted(out, key=lambda y: y[1], reverse=True)

def build_datestack(obj, date_field, qs=None):
    """ Build a sorted list of tuples matching a date to the number of dated events on that date including dates with zero """
    if not qs and not obj:
        return []
    if qs:
        obj_list = qs.order_by(date_field)
    else:
        obj_list = obj.objects.all().order_by(date_field)
    first_date = getattr(obj_list[0], date_field)
    last_date = getattr(obj_list[len(obj_list) -1], date_field)
    try:
        first_date = first_date.date()
        first_date = last_date.date()
        date_list = sorted(list_frequency([getattr(x,date_field).date() for x in obj_list]), key=lambda y: y[0], reverse=True)
    except AttributeError:
        date_list = sorted(list_frequency([getattr(x,date_field) for x in obj_list]), key=lambda y: y[0], reverse=True)
        pass
    out = []
    builder_date = first_date
    data_date = date_list.pop()
    while builder_date <= last_date:
        if data_date[0] == builder_date:
            out.append((builder_date, data_date[1]))
            try:
                data_date = date_list.pop()
            except IndexError:
                pass
        else:
            out.append((builder_date, 0))
        builder_date = builder_date + datetime.timedelta(days=1)
    return out

def rolling_average(datestack, days=7):
    """ Build a new datestack based on the one made by build_datestack that computes a rolling average"""
    # Assuming we've already summed everything, sorted everything, and zero days are included
    out = []
    for x in range(0, len(datestack)):
        if x > days:
            # calculate the sum
            sample_set = datestack[x - days:x]
            counter = 0
            for date, count in sample_set:
                counter = counter + count
            out.append((date, counter/float(days)))
    return out

def build_report(report_list, average=None):
    """ Using a list of dictionaries to build datestacks to work with, and optionally a rolling average
    [ dict(
        obj        = User, # A Django model
        date_field = 'date_joined', # the date or datetime field to count
        label      = "Users", # A convenient label to identify what's being counted
    ), ]

    In a view, you might add this to a context like this:
        for key, value in report.iteritems():
            context.update({key:value})

    Or as a serialized json object like this:
        dthandler = lambda obj: time.mktime(obj.timetuple()) if isinstance(obj, datetime.date) else None
        context['report_as_json'] = json.dumps(report, default=dthandler)
    """

    now = datetime.date.today()
    report = {}
    for each in report_list:
        obj = each.get('obj', None)
        date_field = each.get('date_field', None)
        label = each.get('label', None)
        qs = each.get('qs', None)
        if qs is None:
            qs = obj.objects.all()
        result = build_datestack(obj, date_field, qs=qs)
        report[label]  = result
        if average:
            # Check for "True",
            if isinstance(average, bool):
                report['rolling_%s' % label] = rolling_average(result)
            else:
                try:
                    report['rolling_%s'] = rolling_average(result, days=int(average))
                except TypeError:
                    # That's strange.  I expected average to be boolean or an int.  It wasn't  I don't understand what to do
                    pass

        report["%s_all_time" % label] = len(qs)
        try:
            report["%s_last_30" % label] = len([x for x in qs if getattr(x, date_field) > now - datetime.timedelta(days=30)])
            report["%s_last_60" % label] = len([x for x in qs if getattr(x, date_field) > now - datetime.timedelta(days=60)])
            report["%s_last_90" % label] = len([x for x in qs if getattr(x, date_field) > now - datetime.timedelta(days=90)])
        except TypeError: # These are datetimes
            report["%s_last_30" % label] = len([x for x in qs if getattr(x, date_field).date() > now - datetime.timedelta(days=30)])
            report["%s_last_60" % label] = len([x for x in qs if getattr(x, date_field).date() > now - datetime.timedelta(days=60)])
            report["%s_last_90" % label] = len([x for x in qs if getattr(x, date_field).date() > now - datetime.timedelta(days=90)])
    return report




# Create your views here.
from .tasks import record_request
import datetime
from pytz import utc

def LogRequestMixin(object):
    def dispatch(self, request, *args, **kwargs):
        record_request.delay(
            path = request.path,
            user = request.user,
            http_user_agent = request.META.get('HTTP_USER_AGENT'),
            http_referer = request.META.get('HTTP_REFERER'),
            remote_addr = request.META.get('REMOTE_ADDR'),
            server_name = request.META.get('SERVER_NAME'),
            timestamp = datetime.datetime.utcnow().replace(tzinfo=utc)
        )
        return super(LogRequestMixin, self).dispatch(request, *args, **kwargs)



from django.db import models
from django.conf import settings

class Metric(models.Model):
    name = models.CharField(max_length=64, blank=True, null=True)
    description = models.CharField(max_length=64, blank=True, null=True)
    count = models.IntegerField()

    def increment(self):
        self.count = self.count + 1

class DayMetric(Metric):
    when = models.DateField(null=True, blank=True)

class WeekMetric(Metric):
    when = models.DateField(null=True, blank=True)

class MonthMetric(Metric):
    when = models.DateField(null=True, blank=True)



from .models import tx_add_request
try:
    from celery.task import task
except ImportError:
    from celery.decorators import task


@task
def record_request(path=None,
                   user=None,
                   http_user_agent=None,
                   http_referer=None,
                   remote_addr=None,
                   server_name=None,
                   timestamp=None):
    tx_add_request(path = path,
                   user = user,
                   http_user_agent = http_user_agent,
                   http_referer = http_referer,
                   remote_addr = remote_addr,
                   server_name = server_name,
                   timestamp = timestamp)




from .tasks import record_request
import datetime
from pytz import utc

class RecordRequestMiddleware(object):
    def process_template_response(self, request, response):
        if response.status_code == 200:
            try:
                # grab things from the request object
                record_request.delay(
                    path = request.path,
                    user = request.user,
                    http_user_agent = request.META.get('HTTP_USER_AGENT'),
                    http_referer = request.META.get('HTTP_REFERER'),
                    remote_addr = request.META.get('REMOTE_ADDR'),
                    server_name = request.META.get('SERVER_NAME'),
                    timestamp = datetime.datetime.utcnow().replace(tzinfo=utc)
                )
            except Exception:
                pass
        return response



from django.db import models
from django.conf import settings

class Event(models.Model):
    who = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True)
    recorded_time = models.DateTimeField(auto_now_add=True)
    when = models.DateTimeField(null=True, blank=True)
    category = models.CharField(max_length=64, blank=True, null=True)
    action = models.CharField(max_length=64, blank=True, null=True)
    label = models.CharField(max_length=64, blank=True, null=True)
    what = models.CharField(max_length=255, blank=True, null=True)





from logging import getLogger
logger = getLogger('telemetry_requests')
import datetime
from django.utils.timezone import utc
from django.db import models, transaction
from django.conf import settings

# Create your models here.

class UserAgent(models.Model):
    http_user_agent = models.CharField(max_length=255, unique=True)
    is_spider = models.BooleanField(default=False)
    first_seen = models.DateTimeField(auto_now_add=True)

    def __unicode__(self):
        return self.http_user_agent[:20]

    @property
    def request_count(self):
        return len(self.requests_set.all())


class HttpReferer(models.Model):
    http_referer = models.URLField(max_length=255, unique=True)
    first_seen = models.DateTimeField(auto_now_add=True)

    @property
    def request_count(self):
        return len(self.requests_set.all())

    def __unicode__(self):
        return self.http_referer[:20]



class LiveRequestManager(models.Manager):
    def get_queryset(self):
        return super(LiveRequestManager, self).get_queryset().filter(http_referer__is_spider=False)


class SpiderRequestManager(models.Manager):
    def get_queryset(self):
        return super(SpiderRequestManager, self).get_queryset().filter(http_referer__is_spider=True)


class Request(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, default=None)
    http_user_agent = models.ForeignKey(UserAgent, null=True, default=None)
    http_referer = models.ForeignKey(HttpReferer, null=True, default=None)
    timestamp = models.DateTimeField(auto_now_add=True)
    path = models.CharField(max_length=256)
    remote_addr = models.CharField(max_length=256)
    server_name = models.CharField(max_length=256)
    by_people = LiveRequestManager()
    by_spider = SpiderRequestManager()


def add_request(request):
    # grab things from the request object
    tx_add_request(
        path = request.path,
        user = request.user,
        http_user_agent = request.META.get('HTTP_USER_AGENT'),
        http_referer = request.META.get('HTTP_REFERER'),
        remote_addr = request.META.get('REMOTE_ADDR'),
        server_name = request.META.get('SERVER_NAME'),
        timestamp = datetime.datetime.utcnow().replace(tzinfo=utc)
    )



@transaction.commit_on_success
def tx_add_request(path=None,
                   user=None,
                   http_user_agent=None,
                   http_referer=None,
                   remote_addr=None,
                   server_name=None,
                   timestamp=None):
    """Made to be run within a transaction so that it all succeeds or fails together """
    if not path:
        return None
    if http_user_agent:
        normalized = http_user_agent.strip()[:255]
        useragent, created = UserAgent.objects.get_or_create(http_user_agent=normalized)
    else:
        useragent = None
    if  http_referer:
        normalized = http_referer.strip()[:255]
        referer, created = HttpReferer.objects.get_or_create(http_referer=normalized)
    else:
        referer = None


    create_dict = dict(
        http_user_agent = useragent,
        http_referer = referer,
        path = path
    )
    if user:
        if user.is_authenticated():
            create_dict.update(user = user)

    if remote_addr:
        create_dict.update(dict(
            remote_addr = remote_addr
        ))
    if server_name:
        create_dict.update(dict(
            server_name = server_name
        ))
    if timestamp:
        create_dict.update(dict(
            timestamp = timestamp
        ))
    request_obj =  Request(**create_dict)
    request_obj.save()


def log_request(request):
    if request.user.is_authenticated():
        logger.info('%s viewed %s' % (request.user.email, request.path))




#import logging
#logger = logging.getLogger(__name__)
from .models import Event
from django.contrib.auth import get_user_model
User = get_user_model()

try:
    from celery.task import task
except ImportError:
    from celery.decorators import task

@task
def add_metric( who=None,
                when=None,
                what=None,
                category=None,
                action=None,
                label=None
               ):
        if who:
            #assume it's an e-mail
            try:
                user = User.objects.get(email=who)
            except User.DoesNotExist:
                user = None
        else:
            user = None
        obj = Event(
            who = user,
            when = when,
            what = unicode(what),
            category = unicode(category),
            action = unicode(action),
            label = unicode(label),
        )
        #logger.debug("prepared to record a metric %s " % what)
        obj.save()



from django.views.generic import TemplateView
from .utils import build_report
import datetime
import time
#TODO change this to the pluggable user models
import json
from django.contrib.auth import get_user_model

User = get_user_model()

class UserGraphView(TemplateView):
    template_name = "telemetry/usergraph.html"
    def get_context_data(self, **kwargs):
        context = super(TemplateView, self).get_context_data(**kwargs)
        report = build_report([
            dict(obj=User,date_field='date_joined',label="Users"),
        ], average=True)
        # This is naive and only serves as an example
        dthandler = lambda obj: time.mktime(obj.timetuple()) if isinstance(obj, datetime.date) else None
        context.update({'totals_json': json.dumps(report, default=dthandler), })
        for key, value in report.iteritems():
            context.update({key: value})
        return context



import logging
logger = logging.getLogger(__name__)
from django.contrib import admin
from .models import (
    Request,
    UserAgent,
    HttpReferer
)

class HttpRefererAdmin(admin.ModelAdmin):
    list_display = ('http_referer', 'first_seen')

class UserAgentAdmin(admin.ModelAdmin):
    list_display = ('http_user_agent', 'first_seen', 'is_spider')

class RequestAdmin(admin.ModelAdmin):
    list_filter = ('path', 'user')
    search_fields = ('path', 'user')
    list_display = ('server_name', 'path', 'remote_addr', 'user', 'http_user_agent', 'http_referer')

admin.site.register(Request, RequestAdmin)
admin.site.register(UserAgent, UserAgentAdmin)
admin.site.register(HttpReferer, HttpRefererAdmin)



from distutils.core import setup
import os


packages, data_files = [], []
root_dir = os.path.dirname(__file__)
if root_dir:
    os.chdir(root_dir)

for dirpath, dirnames, filenames in os.walk('telemetry'):
    # Ignore dirnames that start with '.'
    for i, dirname in enumerate(dirnames):
        if dirname.startswith('.'): del dirnames[i]
    if '__init__.py' in filenames:
        pkg = dirpath.replace(os.path.sep, '.')
        if os.path.altsep:
            pkg = pkg.replace(os.path.altsep, '.')
        packages.append(pkg)
    elif filenames:
        prefix = dirpath
        for f in filenames:
            data_files.append(os.path.join(prefix, f))
print data_files

setup(
    name='django-telemetry',
    version='0.1.1',
    author='Alex Lovell-Troy',
    author_email='alex@lovelltroy.org',
    description='Event-based reporting system for django - with queryset graphs',
    packages=[
        'telemetry',
        'telemetry.events',
        'telemetry.events.migrations',
        'telemetry.metrics',
        'telemetry.reporting',
        'telemetry.requests',
    ],
    include_package_data=True,
    package_data = { 'telemetry.reporting': [
        "static/javascripts/*.js",
        "static/*.css",
        "templates/*.html",
    ]},
    url='https://github.com/alexlovelltroy/django-telemetry',
    license='LICENSE.txt',
    long_description=open('README.md').read(),
    install_requires=[
        "Django >= 1.5",
        "pytz >= 2013.9",
        "django-celery >= 3.1",
        "South >= 0.8",
    ],
)



from django.contrib import admin

from .models import Event

class EventAdmin(admin.ModelAdmin):
    list_display = (
                    'recorded_time',
                    'who',
                    'category',
                    'action',
                    'label',
                    'what',
                    'when',
    )

admin.site.register(Event, EventAdmin)
